/* $LAAS: api.c 2011/05/18 17:54:10 mallet $ */

/*
 * Copyright (c) 2008-2009,2011 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Fri Jul  4 2008
 */

#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "evart-client.h"


/* --- local data ---------------------------------------------------- */

static unsigned short sport =		EVAS_STREAM_PORT;
static char shost[1024] =		EVAS_STREAM_HOST;
struct sockaddr_in saddr;
static int ssockfd =			-1;

static int (*whandler)(const evas_msg_t *, evas_msgtype_t, void *) = NULL;
static evas_msgtype_t wtype = EVAS_NONE;
static void *wdata = NULL;

static int (*shandler)(const evas_msg_t *, void *) = NULL;
static void *sdata = NULL;


static int			evas_send(const evas_msg_t *msg);
static const evas_msg_t *	evas_wait(const evas_msg_t *src,
					evas_msg_t *reply, evas_msgtype_t type);
static int			evas_wait_handler(const evas_msg_t *msg,
					evas_msgtype_t type, void *data);


/* --- evas_setport -------------------------------------------------- */

/** \ingroup init
 * \brief Configure the IP port used by the evart-stream server
 *
 * This function can be called at anytime, but it's probably more useful
 * to call it before any function that contact the remote server (see
 * \ref data). In any case, subsequent calls to such functions will use
 * the new \a port value.
 *
 * \param[in] port	The port number, as configured on the remote
 *			server.
 */

void
evas_setport(uint16_t port)
{
  sport = port;

  if (ssockfd >= 0) {
    close(ssockfd);
    ssockfd = -1;
  }
}


/* --- evas_sethost -------------------------------------------------- */

/** \ingroup init
 * \brief Configure the host name where the evart-stream server is
 * running.
 *
 * This function can be called at anytime, but it's probably more useful
 * to call it before any function that contact the remote server (see
 * \ref data). In any case, subsequent calls to such functions will use
 * the new \a host value.
 *
 * \param[in] host	The host (IP address or name) where the remote
 *			server is running. If \a host is NULL, then the
 *			local host is used.
 */

void
evas_sethost(const char *host)
{
  strncpy(shost, host?host:"localhost", sizeof(shost)-1);

  if (ssockfd >= 0) {
    close(ssockfd);
    ssockfd = -1;
  }
}


/* --- evas_sethandler ----------------------------------------------- */

/** \ingroup data
 * \brief Register a callback function to be invoked when new data is
 *        received.
 *
 * Use this function in conjunction with evas_listen() to process the
 * data from the server. The \a handler() callback function is called
 * each time a new message is received, provided that the evas_listen()
 * function is being executed. \a handler() is also called if an
 * irrelevant message is received when a blocking function of the API is
 * waiting for an answer, so this is the preferred mechanism for
 * receiving data since no message can be lost.
 *
 * The \a handler() function should normally return zero. If it returns
 * some other value, evas_listen() will stop listening for data and
 * return immediately.
 *
 * \param[in] handler	The callback function. If NULL, the callback
 *                      mechanism is disabled.
 * \param[in] data	Pointer to user data that gets passed to each
 *                      invocation of \a handler.
 */

void
evas_sethandler(int (*handler)(const evas_msg_t *, void *), void *data)
{
  sdata = data;
  shandler = handler;
}


/* --- evas_acquire -------------------------------------------------- */

/** \ingroup control
 * \brief Start or pause marker tracking in EVaRT software.
 *
 * Calling this function will ask the EVaRT software to track the marker
 * sets. evas_acquire() will block until it gets the confirmation from
 * the server. If other messages are received in the meantime and a \em
 * callback has been defined (see evas_sethandler()), the callback is
 * invoked with those messages.
 *
 * \param[in] on	Use EVAS_ON for starting the tracking and
 *			EVAS_OFF for pausing it.
 *
 * \return 0 on success, anything else in case of error.
 */

int
evas_acquire(evas_onoff_t on)
{
  evas_msg_t r;
  const evas_msg_t *a;
  int n;

  /* create message */
  evas_msg_t msg = { .type = EVAS_ACQUIRE };
  n = snprintf(msg.cmd, sizeof(msg.cmd), "%c", on?'+':'-');
  msg.len = (char *)&msg.cmd[n+1] - (char *)&msg;

  /* wait for reply */
  a = evas_wait(&msg, &r, EVAS_ACQUIRE);
  if (!a) return -1;

  /* return data */
  return atoi(a->cmd);
}


/* --- evas_record --------------------------------------------------- */

/** \ingroup control
 * \brief Start or stop marker data recording in EVaRT software.
 *
 * This function asks the EVaRT software to dump the marker
 * sets on a file. evas_record() will block until it gets the
 * confirmation from the server. If other messages are received in the
 * meantime and a \em callback has been defined (see evas_sethandler()),
 * the callback is invoked with those messages.
 *
 * \param[in] file	Filename in which to save the data on the remote
 *			host. \a file should not contain any path
 *			elements (i.e. this is a pure filename with no
 *			directories). \a file is not used (can be NULL)
 *			if \a on is EVAS_OFF.
 * \param[in] on	Use EVAS_ON for starting the recording and
 *			EVAS_OFF for stopping it.
 *
 * \return 0 on success, anything else in case of error.
 */

int
evas_record(const char *file, evas_onoff_t on)
{
  evas_msg_t r;
  const evas_msg_t *a;
  int n;

  /* create message */
  evas_msg_t msg = { .type = EVAS_RECORD };
  n = snprintf(msg.cmd, sizeof(msg.cmd), "%s", on?file:"");
  msg.len = (char *)&msg.cmd[n+1] - (char *)&msg;

  /* wait for reply */
  a = evas_wait(&msg, &r, EVAS_RECORD);
  if (!a) return -1;

  /* return data */
  return atoi(a->cmd);
}


/* --- evas_fps ------------------------------------------------------ */

/** \ingroup data
 * \brief Get marker data frame rate.
 *
 * This function asks the EVaRT software for the marker data frame
 * rate. evas_fps() will block until it gets the response from the
 * evas-stream server. If other messages are received in the meantime and
 * a \em  callback has been defined (see evas_sethandler()), the callback
 * is invoked with those messages.
 *
 * \param[out] fps	Frame rate value, in frames per second.
 *
 * \return 0 on success, anything else in case of error.
 */

int
evas_fps(float *fps)
{
  evas_msg_t r;
  const evas_msg_t *a;

  /* compute message */
  evas_msg_t msg = { .type = EVAS_FPS };
  msg.len = (char *)&msg.cmd[0] - (char *)&msg;

  /* wait for reply */
  a = evas_wait(&msg, &r, EVAS_FPS);
  if (!a) return -1;

  /* return data */
  if (fps) *fps = atof(a->cmd);
  return a->cmd[0] == '+'?0:-1;
}


/* --- evas_body_list ------------------------------------------------ */

/** \ingroup requests
 * \brief Get the current list of tracked bodies (a.k.a. "markersets").
 *
 * Calling this function will ask the server to transmit the current list
 * of tracked bodies ("markersets" in the EVaRT(tm) GUI) and will block
 * until it receives the response. If other messages are received in the
 * meantime and a \em callback has been defined (see evas_sethandler()),
 * the callback is invoked with those messages.
 *
 * \return A pointer on static storage containing the current list of
 *         bodies, or NULL in case of error. See evas_body_list_t.
 */

const evas_body_list_t *
evas_body_list()
{
  static evas_msg_t r;
  const evas_msg_t *a;

  /* create message */
  evas_msg_t msg = { .type = EVAS_BODY_LIST };
  msg.len = (char *)&msg.cmd[0] - (char *)&msg;

  /* wait for reply */
  a = evas_wait(&msg, &r, EVAS_BODY_LIST);
  if (!a) return NULL;

  /* return data */
  return &a->body_list;
}


/* --- evas_body_markers_list ---------------------------------------- */

/** \ingroup requests
 * \brief Get the markers list for one body.
 *
 * Calling this function will ask the server to transmit the markers list
 * of the specified \a body. It will block until it receives the
 * response. If other messages are received in the meantime and a \em
 * callback has been defined (see evas_sethandler()), the callback is
 * invoked with those messages.
 *
 * \param[in] body	A valid body index. In order to find out which \a
 *			body indexes are available, you might call
 *			evas_body_list() beforehand.
 *
 * \return A pointer on static storage containing the list of markers
 *         of the \a body, or NULL in case of error. See
 *         evas_body_markers_list_t.
 */

const evas_body_markers_list_t *
evas_body_markers_list(uint32_t body)
{
  static evas_msg_t r;
  const evas_msg_t *a;
  int n;

  /* create message */
  evas_msg_t msg = { .type = EVAS_BODY_MARKERS_LIST };
  n = snprintf(msg.cmd, sizeof(msg.cmd), "%d", body);
  msg.len = (char *)&msg.cmd[n+1] - (char *)&msg;

  /* wait for reply */
  do {
    a = evas_wait(&msg, &r, EVAS_BODY_MARKERS_LIST);
    if (!a) return NULL;
    if (a->body_markers_list.index != body) {
      printf("* skipping marker list for body %d\n",
	     a->body_markers_list.index);
      if (shandler) shandler(a, sdata);
    }
  } while(a->body_markers_list.index != body);

  /* return data */
  return &a->body_markers_list;
}


/* --- evas_body_segments_list --------------------------------------------- */

/** \ingroup requests
 * \brief Get the segments list for one body.
 *
 * Calling this function will ask the server to transmit the segments list
 * of the specified \a body. It will block until it receives the
 * response. If other messages are received in the meantime and a \em
 * callback has been defined (see evas_sethandler()), the callback is
 * invoked with those messages.
 *
 * \param[in] body	A valid body index. In order to find out which \a
 *			body indexes are available, you might call
 *			evas_body_list() beforehand.
 *
 * \return A pointer on static storage containing the list of segments
 *         of the \a body, or NULL in case of error. See
 *         evas_body_segments_list_t.
 */

const evas_body_segments_list_t *
evas_body_segments_list(uint32_t body)
{
  static evas_msg_t r;
  const evas_msg_t *a;
  int n;

  /* create message */
  evas_msg_t msg = { .type = EVAS_BODY_SEGMENTS_LIST };
  n = snprintf(msg.cmd, sizeof(msg.cmd), "%d", body);
  msg.len = (char *)&msg.cmd[n+1] - (char *)&msg;

  /* wait for reply */
  do {
    a = evas_wait(&msg, &r, EVAS_BODY_SEGMENTS_LIST);
    if (!a) return NULL;
    if (a->body_segments_list.index != body) {
      printf("* skipping segments list for body %d\n",
	     a->body_segments_list.index);
      if (shandler) shandler(a, sdata);
    }
  } while(a->body_segments_list.index != body);

  /* return data */
  return &a->body_segments_list;
}


/* --- evas_body_dofs_list --------------------------------------------- */

/** \ingroup requests
 * \brief Get the dofs list for one body.
 *
 * Calling this function will ask the server to transmit the dofs list
 * of the specified \a body. It will block until it receives the
 * response. If other messages are received in the meantime and a \em
 * callback has been defined (see evas_sethandler()), the callback is
 * invoked with those messages.
 *
 * \param[in] body	A valid body index. In order to find out which \a
 *			body indexes are available, you might call
 *			evas_body_list() beforehand.
 *
 * \return A pointer on static storage containing the list of dofs
 *         of the \a body, or NULL in case of error. See
 *         evas_body_dofs_list_t.
 */

const evas_body_dofs_list_t *
evas_body_dofs_list(uint32_t body)
{
  static evas_msg_t r;
  const evas_msg_t *a;
  int n;

  /* create message */
  evas_msg_t msg = { .type = EVAS_BODY_DOFS_LIST };
  n = snprintf(msg.cmd, sizeof(msg.cmd), "%d", body);
  msg.len = (char *)&msg.cmd[n+1] - (char *)&msg;

  /* wait for reply */
  do {
    a = evas_wait(&msg, &r, EVAS_BODY_DOFS_LIST);
    if (!a) return NULL;
    if (a->body_dofs_list.index != body) {
      printf("* skipping dofs list for body %d\n",
	     a->body_dofs_list.index);
      if (shandler) shandler(a, sdata);
    }
  } while(a->body_dofs_list.index != body);

  /* return data */
  return &a->body_dofs_list;
}


/* --- evas_body_markers --------------------------------------------- */

/** \ingroup requests
 * \brief Configure the streaming of markers positions for one body.
 *
 * Calling this function will configure the server so that it will start
 * streaming marker positions for the given \a body. The function will
 * return immediately and you should be prepared to receive messages of
 * the type EVAS_BODY_MARKERS in your \em callback function (see
 * evas_sethandler()) or in evas_recv().
 *
 * \param[in] body	A valid body index. In order to find out which \a
 *			body indexes are available, you might call
 *			evas_body_list() beforehand.
 * \param[in] on	A boolean that starts (EVAS_ON) or stops
 *                      (EVAS_OFF) the streaming of marker data for this
 *                      \a body.
 *
 * \return 0 on success, anything else if an error occur.
 */

int
evas_body_markers(uint32_t body, evas_onoff_t on)
{
  evas_msg_t msg = { .type = EVAS_BODY_MARKERS };
  int n;

  n = snprintf(msg.cmd, sizeof(msg.cmd), "%c%d ka", on?'+':'-', body);
  msg.len = (char *)&msg.cmd[n+1] - (char *)&msg;

  return evas_send(&msg);
}


/* --- evas_body_segments -------------------------------------------------- */

/** \ingroup requests
 * \brief Configure the streaming of segments positions for one body.
 *
 * Calling this function will configure the server so that it will start
 * streaming segments positions for the given \a body. The function will
 * return immediately and you should be prepared to receive messages of
 * the type EVAS_BODY_SEGMENTS in your \em callback function (see
 * evas_sethandler()) or in evas_recv().
 *
 * \param[in] body	A valid body index. In order to find out which \a
 *			body indexes are available, you might call
 *			evas_body_list() beforehand.
 * \param[in] on	A boolean that starts (EVAS_ON) or stops
 *                      (EVAS_OFF) the streaming of segment data for this
 *                      \a body.
 *
 * \return 0 on success, anything else if an error occur.
 */

int
evas_body_segments(uint32_t body, evas_onoff_t on)
{
  evas_msg_t msg = { .type = EVAS_BODY_SEGMENTS };
  int n;

  n = snprintf(msg.cmd, sizeof(msg.cmd), "%c%d ka", on?'+':'-', body);
  msg.len = (char *)&msg.cmd[n+1] - (char *)&msg;

  return evas_send(&msg);
}


/* --- evas_body_dofs --------------------------------------------- */

/** \ingroup requests
 * \brief Configure the streaming of DOFs positions for one body.
 *
 * Calling this function will configure the server so that it will start
 * streaming DOFs positions for the given \a body. The function will
 * return immediately and you should be prepared to receive messages of
 * the type EVAS_BODY_DOFS in your \em callback function (see
 * evas_sethandler()) or in evas_recv().
 *
 * \param[in] body	A valid body index. In order to find out which \a
 *			body indexes are available, you might call
 *			evas_body_list() beforehand.
 * \param[in] on	A boolean that starts (EVAS_ON) or stops
 *                      (EVAS_OFF) the streaming of DOF data for this
 *                      \a body.
 *
 * \return 0 on success, anything else if an error occur.
 */

int
evas_body_dofs(uint32_t body, evas_onoff_t on)
{
  evas_msg_t msg = { .type = EVAS_BODY_DOFS };
  int n;

  n = snprintf(msg.cmd, sizeof(msg.cmd), "%c%d ka", on?'+':'-', body);
  msg.len = (char *)&msg.cmd[n+1] - (char *)&msg;

  return evas_send(&msg);
}


/* --- evas_unrecognized_markers ------------------------------------- */

/** \ingroup requests
 * \brief Configure the streaming of unrecognized markers positions.
 *
 * Calling this function will configure the server so that it will start
 * streaming unrecognized marker positions. Unrecognized markers are
 * those that do not belong to any body (see evas_body_markers()). The
 * function will return immediately and you should be prepared to receive
 * messages of the type EVAS_UNRECOGNIZED_MARKERS in your \em callback
 * function (see evas_sethandler()) or in evas_recv().
 *
 * \param[in] on	A boolean that starts (EVAS_ON) or stops
 *                      (EVAS_OFF) the streaming of unrecognized marker
 *                      data.
 *
 * \return 0 on success, anything else if an error occur.
 */

int
evas_unrecognized_markers(evas_onoff_t on)
{
  evas_msg_t msg = { .type = EVAS_UNRECOGNIZED_MARKERS };
  int n;

  n = snprintf(msg.cmd, sizeof(msg.cmd), "%c ka", on?'+':'-');
  msg.len = (char *)&msg.cmd[n+1] - (char *)&msg;

  return evas_send(&msg);
}


/* --- evas_send ----------------------------------------------------- */

static int
evas_send(const evas_msg_t *msg)
{
  ssize_t sz;

  if (ssockfd < 0) {
    char hport[6];
    struct addrinfo *haddr, *a;
    int reuse;
    int s;

    ssockfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (ssockfd == -1) return -1;

    /* set reuse addr option */
    reuse = 1;
    if (setsockopt(ssockfd, SOL_SOCKET, SO_REUSEADDR,
		   (char *)&reuse, sizeof(reuse))) {
      warnx("cannot set reuseaddr on server socket");
      close(ssockfd);
      ssockfd = -1;
      return -1;
    }

    /* resolve host address */
    sprintf(hport, "%d", sport);
    s = getaddrinfo(shost, hport, NULL, &haddr);
    if (s) { warnx("cannot get remote host addresses"); return s; }

    saddr.sin_addr.s_addr = htonl(INADDR_ANY);
    for(a = haddr; a; a = a->ai_next) switch (a->ai_family) {
      case AF_INET:
	saddr = *(struct sockaddr_in *)a->ai_addr;
	break;

      default: break;
    }
    if (saddr.sin_addr.s_addr == htonl(INADDR_ANY)) {
      warnx("cannot resolve remote host address");
      close(ssockfd);
      ssockfd = -1;
      return -1;
    }
  }

  sz = sendto(ssockfd, (const char *)msg, msg->len, 0,
	      (struct sockaddr *)&saddr, sizeof(saddr));
  if (sz != msg->len) return -1;

  return 0;
}


/* --- evas_text ----------------------------------------------------- */

/** \ingroup requests
 * \brief Miscellaneous messages.
 *
 * This function sends a pure text message to the server. The behaviour
 * depends on the actual message sent. This function does not wait for
 * any answer. If your message generates one, you will receive it
 * asynchronously.
 *
 * At the moment, only one message is recognized by the server: it is the
 * ASCII text "hello" terminated by a nul-character. The server will
 * respond with a text message (type EVAS_MSG) containing its
 * identification string. You can use this message to test the
 * communication with the server.
 *
 * \param[in] text	Raw text message to send to the server. At the
 *                      moment, only the nul-terminated string "hello" is
 *                      useful.
 *
 * \return 0 on success, anything else if an error occur.
 */

int
evas_text(const char *text)
{
  evas_msg_t msg = { .type = EVAS_MSG };
  int n;

  n = snprintf(msg.cmd, sizeof(msg.cmd), "%s", text);
  msg.len = (char *)&msg.cmd[n+1] - (char *)&msg;

  return evas_send(&msg);
}


/* --- evas_wait ----------------------------------------------------- */

static const evas_msg_t *
evas_wait(const evas_msg_t *src, evas_msg_t *reply, evas_msgtype_t type)
{
  evas_msg_t msg;
  const evas_msg_t *a;
  struct timeval now, end;
  int retries = 3;

  /* setup wait handler (in case another evas_recv() is running) */
  msg.type = EVAS_NONE;
  wdata = &msg;
  wtype = type;
  whandler = evas_wait_handler;

  /* compute absolute deadline (1s later) */
  gettimeofday(&end, NULL);
  end.tv_sec++;

  /* loop on replies */
  do {
    /* send message */
    if (evas_send(src)) return NULL;

    for(a = evas_recv(reply, 1.); a; a = evas_recv(reply, 1.)) {
      if (reply->type == type) { whandler = NULL; return reply; }
      if (msg.type == type) { *reply = msg; return reply; }
      if (shandler) shandler(a, sdata);

      /* check the deadline */
      gettimeofday(&now, NULL);
      if (now.tv_sec > end.tv_sec ||
	  (now.tv_sec == end.tv_sec && now.tv_usec > end.tv_usec))
	break;
    }

    /* timeout */
    if (msg.type == type) { *reply = msg; return reply; }

    /* compute new deadline */
    gettimeofday(&end, NULL);
    end.tv_sec++;
  } while(--retries >= 0);

  whandler = NULL;
  return NULL;
}

static int
evas_wait_handler(const evas_msg_t *msg, evas_msgtype_t type, void *data)
{
  evas_msg_t *wmsg = data;

  if (msg->type == type) {
    whandler = NULL;
    *wmsg = *msg;
  }
  return 0;
}


/* --- evas_listen --------------------------------------------------- */

/** \ingroup data
 * \brief Block and process received data.
 *
 * This function blocks on message reception infinitely. It is useless to
 * call this function if you do not have setup a \em callback (with
 * evas_sethandler()) beforehand.
 *
 * As soon as a message is received, the registered callback (if any) is
 * invoked with the message. If the \a callbak function returns anything
 * else than 0, the evas_listen() function returns immediately after
 * this. Otherwise, evas_listen() waits for the next message and loops
 * forever.
 */

void
evas_listen()
{
  evas_msg_t msg;
  const evas_msg_t *a;

  for(a = evas_recv(&msg, 0.); a; a = evas_recv(&msg, 0.)) {
    if (whandler) whandler(a, wtype, wdata);
    if (shandler) if (shandler(a, sdata)) break;
  }
}


/* --- evas_recv ----------------------------------------------------- */

/** \ingroup data
 * \brief Poll for some data.
 *
 * This function blocks until a message is received and returns that
 * message. Any registered \em callback (with evas_sethandler()) is \b
 * not called.
 *
 * You can use this function if you want to make your own event
 * processing (for instance if you have several event channels to listen
 * for).
 *
 * \param[in] msg	Pointer on a user-allocated message, filled by the
 *			function.
 * \param[in] timeout	Maximum time in seconds to wait for a message, or 0 to
 *			wait forever.
 *
 * \return A pointer to the received message, or NULL in case of error.
 */

const evas_msg_t *
evas_recv(evas_msg_t *msg, double timeout)
{
  struct sockaddr from;
  size_t fromlen = sizeof(from);
  struct timeval maxwait;
  fd_set fds;
  int s;

  if (ssockfd < 0) return NULL;

  do {
    if (timeout > 0.) {
      do {
        maxwait.tv_sec = timeout;
        maxwait.tv_usec = (timeout - maxwait.tv_sec) * 1e6;
        FD_ZERO(&fds);
        FD_SET(ssockfd, &fds);
        s = select(ssockfd+1, &fds, NULL, NULL, &maxwait);
      } while(s == -1 && errno == EAGAIN);
      if (s <= 0) return NULL;
    }

    do {
      s = recvfrom(ssockfd, msg, sizeof(*msg), 0, &from, &fromlen);
    } while(s == -1 && errno == EAGAIN);
    if (s <= 0) return NULL;

    /* respond to keep alive requests and wait for next message */
    if (msg->type == EVAS_KEEPALIVE) (void)evas_send(msg);
  } while(msg->type == EVAS_KEEPALIVE);

  if (whandler) whandler(msg, wtype, wdata);
  return msg;
}
